Require Import coherent.
Require Import committable.
Require Import Coq.Lists.List.
Require Import Coq.Sets.Ensembles.
Require Import syntax.
Require Import trails.

Inductive global_heap : Set :=
| unfrozen : heap -> global_heap
| frozen : heap -> global_heap.

Definition search_heap := prod heap heap_path.

Definition susp_thread := prod (prod tid ctxt) exp.

Definition local_heap := option search_heap.

Definition search_thread := 
  prod (prod (prod (prod tid local_heap) (list ctxt)) exp) path.

Definition empty_susp := Empty_set susp_thread.
Definition empty_heap := Empty_set search_heap.
Definition empty_search := Empty_set search_thread.

(* nonces - needed for ll2hl proof.  basically you tag each low-level
communication or thread run with a unique ID *)
Axiom get_fresh_nonce : Ensemble search_heap ->
  Ensemble search_thread -> nonce.

(* some functions to extract nonces from paths and search
threads/heaps *)
Fixpoint nonces_path (p:path) : Ensemble nonce :=
  match p with 
    | path_nil => Empty_set nonce
    | path_cons elt p' =>
      match elt with
        | path_left => nonces_path p'
        | path_right => nonces_path p'
        | path_send n_x id_x p_x => 
          Add nonce (Union nonce (nonces_path p_x) (nonces_path p')) n_x
        | path_recv n_x id_x p_x => 
          Add nonce (Union nonce (nonces_path p_x) (nonces_path p')) n_x
        | path_heap n_x hp_x => 
          Add nonce (Union nonce (nonces_heap_path hp_x) (nonces_path p')) n_x
      end
  end
with nonces_heap_path (hp:heap_path) : Ensemble nonce :=
  match hp with
    | heap_path_nil => Empty_set nonce
    | heap_path_cons elt hp' =>
      match elt with
        | heap_path_thread n_x id_x p_x =>
          (Add nonce (Union nonce (nonces_path p_x) 
            (nonces_heap_path hp')) n_x)
      end
  end.

Definition nonces_search_thread (s:search_thread) : Ensemble nonce :=
  match s with
    | (id, h, list_M, e, p) =>
      match h with
        | None => nonces_path p
        | Some (H, hp) => Union nonce 
          (nonces_path p) (nonces_heap_path hp)
    end
  end.

Definition nonces_search_heap (r:search_heap) : Ensemble nonce :=
  match r with
    | (H, hp) => nonces_heap_path hp
  end.

Definition fresh_nonce (S_H:Ensemble search_heap)
  (S:Ensemble search_thread) (n:nonce) : Prop  :=
  (forall r n',
    In search_heap S_H r ->
    In nonce (nonces_search_heap r) n' ->
    n <> n')
  /\ (forall s n',
    In search_thread S s ->
    In nonce (nonces_search_thread s) n' ->
    n <> n').

Axiom get_fresh_nonce_correct :
  forall S_H S,
    fresh_nonce S_H S (get_fresh_nonce S_H S).

Inductive get_S_susp_init (T:Ensemble mt_thread) : Ensemble susp_thread :=
| In_S_susp_init : forall id e M v,
  In mt_thread T (id, e) ->
  get_ctxt e M (sync v) ->
  In susp_thread (get_S_susp_init T) (id, M, v).

Inductive get_S_init (T:Ensemble mt_thread) : Ensemble search_thread :=
| In_S_init : forall id e M v,
  In mt_thread T (id, e) ->
  get_ctxt e M (sync v) ->
  In search_thread (get_S_init T) (id, None, nil, v, path_nil).

Inductive get_thread_trails (S:Ensemble search_thread) 
  : Ensemble trail :=
| In_thread_trails : forall id h list_M e p,
  In search_thread S (id, h, list_M, e, p) ->
  In trail (get_thread_trails S) (thread_trail id p).

Inductive get_T'_complete (S:Ensemble search_thread)
  (S_susp:Ensemble susp_thread) : Ensemble mt_thread :=
| In_T'_complete : forall id M v v' p,
  In search_thread S (id, None, nil, alwaysEvt v', p) ->
  In susp_thread S_susp (id, M, v) ->
  In mt_thread (get_T'_complete S S_susp) (id, ctxt_fill_hole M v').

Inductive translate_susp (S_susp : Ensemble susp_thread) : Ensemble mt_thread :=
| In_translate_susp : forall id M v, 
  In susp_thread S_susp (id, M, v) ->
  In mt_thread (translate_susp S_susp) (id, ctxt_fill_hole M (sync v)).

Lemma translate_susp_empty :
  translate_susp (Empty_set susp_thread) = Empty_set mt_thread.

apply Extensionality_Ensembles. split; intros x x_in_one.
inversion x_in_one as [ id M v false x_eq ]. 
inversion false. inversion x_in_one.

Qed.

Inductive get_S_susp_complete (S_susp:Ensemble susp_thread)
  (S_complete:Ensemble search_thread) : Ensemble susp_thread :=
| In_S_susp_complete : forall id M v h list_M e p,
  In susp_thread S_susp (id, M, v) ->
  In search_thread S_complete (id, h, list_M, e, p) ->
  In susp_thread (get_S_susp_complete S_susp S_complete) (id, M, v).

Definition search_unique_tids (S:Ensemble search_thread) : Prop :=
  forall id h1 list_M1 e1 p1 h2 list_M2 e2 p2,
    In search_thread S (id, h1, list_M1, e1, p1) ->
    In search_thread S (id, h2, list_M2, e2, p2) ->
    (id, h1, list_M1, e1, p1) = (id, h2, list_M2, e2, p2).

Inductive ll_step : 
  global_heap -> Ensemble mt_thread -> Ensemble susp_thread ->
  Ensemble search_heap -> Ensemble search_thread ->
  global_heap -> Ensemble mt_thread -> Ensemble susp_thread ->
  Ensemble search_heap -> Ensemble search_thread -> Prop :=
  
| ll_step_mt : forall H T H' T',
  mt_step H T H' T' ->
  ll_step (unfrozen H) T empty_susp empty_heap empty_search
  (unfrozen H') T' empty_susp empty_heap empty_search

| ll_step_sync_init : forall H T T0 T_all S_susp S S_H,
  (forall id e, In mt_thread T (id, e) -> 
    exists M, exists v, get_ctxt e M (sync v)) ->
  T_all = Union mt_thread T T0 ->
  Disjoint mt_thread T T0 ->
  S_susp = get_S_susp_init T ->
  S = get_S_init T ->
  S_H = Singleton search_heap (H, heap_path_nil) ->
  (ll_step (unfrozen H) T_all empty_susp empty_heap empty_search
    (frozen H) T0 S_susp S_H S)

| ll_step_begin_lockdown : forall H id p list_M e H' hp T S_susp S_H S,
  In search_heap S_H (H', hp) ->
  In search_thread S (id, None, list_M, e, p) ->
  heap_coherent (thread_trail id p) (heap_trail hp) ->
  ll_step (frozen H) T S_susp S_H S
  (frozen H) T S_susp S_H
  (Add search_thread S
    (id, Some (H', hp), list_M, e, p))

| ll_step_end_lockdown : forall H id p list_M e H' hp T S_susp S_H S n,
  fresh_nonce S_H (Add search_thread S (id, Some (H', hp), list_M, e, p)) n ->
  ldvalue list_M e ->
  ll_step (frozen H) T S_susp S_H
  (Add search_thread S (id, Some (H', hp), list_M, e, p))
  (frozen H) T S_susp (Add search_heap S_H 
    (H', heap_path_cons (heap_path_thread n id p) hp)) 
  (Add search_thread S (id, None, list_M, e, (path_cons (path_heap n hp) p)))

| ll_step_send_recv : 
  forall H T S_susp S_H S id1 list_M1 e1 E1 (k:channel) v p1 
    id2 list_M2 e2 E2 p2 n,
    fresh_nonce S_H 
    (Union search_thread S
      (Couple search_thread (id1, None, list_M1, e1, p1) 
        (id2, None, list_M2, e2, p2))) n ->
    get_evt_ctxt e1 E1 (sendEvt k v) ->
    get_evt_ctxt e2 E2 (recvEvt k) ->
    coherent (thread_trail id1 p1) (thread_trail id2 p2) ->
    ll_step (frozen H) T S_susp S_H
    (Union search_thread S
      (Couple search_thread (id1, None, list_M1, e1, p1) (id2, None, list_M2, e2, p2)))
    (frozen H) T S_susp S_H
    (Union search_thread S
      (Union search_thread
        (Couple search_thread (id1, None, list_M1, e1, p1) (id2, None, list_M2, e2, p2))
        (Couple search_thread 
          (id1, None, list_M1, evt_fill_hole E1 (alwaysEvt zero), 
            path_cons (path_send n id2 p2) p1) 
          (id2, None, list_M2, evt_fill_hole E2 (alwaysEvt v), 
            path_cons (path_recv n id1 p1) p2))))

| ll_step_choose :
  forall H T S_susp S_H S id s_H list_M e E v1 v2 p,
    get_evt_ctxt e E (chooseEvt v1 v2) ->
    ll_step (frozen H) T S_susp S_H
    (Add search_thread S (id, Some s_H, list_M, e, p))
    (frozen H) T S_susp S_H
    (Union search_thread S
      (Couple search_thread 
        (id, Some s_H, list_M, evt_fill_hole E v1, path_cons path_left p) 
        (id, Some s_H, list_M, evt_fill_hole E v2, path_cons path_right p)))

| ll_step_then_always :
  forall H T S_susp S_H S id s_H list_M e E v1 v2 p,
    get_evt_ctxt e E (thenEvt (alwaysEvt v1) v2) ->
    ll_step (frozen H) T S_susp S_H
    (Add search_thread S (id, Some s_H, list_M, e, p))
    (frozen H) T S_susp S_H
    (Add search_thread S (id, Some s_H, list_M, evt_fill_hole E (app v2 v1), p))

| ll_step_lockdown :
  forall H T S_susp S_H S id H' H'' hp list_M e e' p,
    st_step H' e H'' e' ->
    ll_step (frozen H) T S_susp S_H
    (Add search_thread S (id, Some (H', hp), list_M, e, p))
    (frozen H) T S_susp S_H
    (Add search_thread S (id, Some (H'', hp), list_M, e', p))

| ll_step_nested_sync_init :
  forall H T S_susp S_H S id s_H list_M e M v p,
    get_ctxt e M (sync v) ->
    ll_step (frozen H) T S_susp S_H
    (Add search_thread S (id, Some s_H, list_M, e, p))
    (frozen H) T S_susp S_H
    (Add search_thread S (id, Some s_H, M::list_M, v, p))

| ll_step_nested_sync_complete :
  forall H T S_susp S_H S id s_H list_M M v p,
    value v ->
    ll_step (frozen H) T S_susp S_H
    (Add search_thread S (id, Some s_H, M::list_M, alwaysEvt v, p))
    (frozen H) T S_susp S_H
    (Add search_thread S (id, Some s_H, list_M, ctxt_fill_hole M v, p))

| ll_step_sync_complete : 
  forall H S T0 T_susp T' T_all' S_susp S_susp0 S_susp_all 
    H' hp S_H S_H_all thread_trails trails S0 S_all,
  (forall id h list_M e p,
    In search_thread S (id, h, list_M, e, p) ->
    (h = None /\ list_M = nil /\
      (exists v', value v' /\ e = alwaysEvt v'))) ->
  search_unique_tids S ->
  T' = get_T'_complete S S_susp ->
  S_susp = get_S_susp_complete S_susp_all S ->
  S_susp0 = Setminus susp_thread S_susp_all S_susp ->
  T_susp = translate_susp S_susp0 ->
  S_H_all = Add search_heap S_H (H', hp) ->
  S_all = Union search_thread S S0 ->
  T_all' = Union mt_thread T' (Union mt_thread T0 T_susp) ->
  thread_trails = get_thread_trails S ->
  trails = Add trail thread_trails (heap_trail hp) ->
  committable trails ->
  (ll_step (frozen H) T0 S_susp_all S_H_all S_all 
    (unfrozen H') T_all' empty_susp empty_heap empty_search).

Inductive ll_step_many : 
  global_heap -> Ensemble mt_thread -> Ensemble susp_thread ->
  Ensemble search_heap -> Ensemble search_thread ->
  global_heap -> Ensemble mt_thread -> Ensemble susp_thread ->
  Ensemble search_heap -> Ensemble search_thread -> Prop :=

| ll_step_many_zero : forall h T S_susp S_H S,
  ll_step_many h T S_susp S_H S h T S_susp S_H S

| ll_step_many_nonzero :
  forall h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 h3 T3 S_susp3 S_H3 S3,
    ll_step h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 ->
    ll_step_many h2 T2 S_susp2 S_H2 S2 h3 T3 S_susp3 S_H3 S3 ->
    ll_step_many h1 T1 S_susp1 S_H1 S1 h3 T3 S_susp3 S_H3 S3.

Lemma ll_step_many_reverse :
  forall h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 h3 T3 S_susp3 S_H3 S3,
    ll_step_many h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 ->
    ll_step h2 T2 S_susp2 S_H2 S2 h3 T3 S_susp3 S_H3 S3 ->
    ll_step_many h1 T1 S_susp1 S_H1 S1 h3 T3 S_susp3 S_H3 S3.

intros h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 h3 T3 S_susp3 S_H3 S3 step_many_1_2 step_2_3.

induction step_many_1_2 as [ h2 T2 S_susp2 S_H2 S2 | 
  h1 T1 S_susp1 S_H1 S1 h1' T1' S_susp1' S_H1' S1' h2 T2 S_susp2 S_H2 S2 step_1_1' step_many_1'_2 IH].
eapply ll_step_many_nonzero.
apply step_2_3.
apply ll_step_many_zero.

eapply ll_step_many_nonzero.
apply step_1_1'.
apply IH.
apply step_2_3.

Qed.

Lemma ll_step_many_many :
  forall h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 h3 T3 S_susp3 S_H3 S3,
    ll_step_many h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 ->
    ll_step_many h2 T2 S_susp2 S_H2 S2 h3 T3 S_susp3 S_H3 S3 ->
    ll_step_many h1 T1 S_susp1 S_H1 S1 h3 T3 S_susp3 S_H3 S3.

intros h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 h3 T3 
  S_susp3 S_H3 S3 step_many_1_2 step_many_2_3.

induction step_many_1_2. auto.
eapply ll_step_many_nonzero.
apply H.
apply IHstep_many_1_2. auto.

Qed.
