Require Import Coq.Lists.List.
Require Import Coq.Sets.Ensembles.

Axiom variable : Set.
Axiom location : Set.
Axiom channel : Set.
Axiom constant : Set.

Axiom zero c1 c2 c3 : constant.
Axiom x1 x2 x3 : variable.
Axiom loc1 loc2 loc3 : location.
Axiom k1 k2 k3 : channel.

Inductive exp : Set :=
  const : constant -> exp
| var : variable -> exp
| lam : variable -> exp -> exp
| app : exp -> exp -> exp
| ref : exp -> exp
| get : exp -> exp
| set : exp -> exp -> exp
| loc : location -> exp
| create : exp -> exp
| sync : exp -> exp
| newChan : exp
| chan : channel -> exp
| alwaysEvt : exp -> exp
| neverEvt : exp
| thenEvt : exp -> exp -> exp
| chooseEvt : exp -> exp -> exp
| sendEvt : exp -> exp -> exp
| recvEvt : exp -> exp.

Coercion var : variable >-> exp.
Coercion loc : location >-> exp.
Coercion chan : channel >-> exp.
Coercion const : constant >-> exp.

Inductive value : exp -> Prop :=
  value_const : forall c : constant, value c
| value_lam : forall x e, value (lam x e)
| value_loc : forall l : location, value l
| value_chan : forall k : channel, value k
| value_always : forall v, value v -> value (alwaysEvt v)
| value_never : value neverEvt
| value_send : forall k v, value k -> value v -> value (sendEvt k v)
| value_recv : forall k, value k -> value (recvEvt k)
| value_then : forall v1 v2, value v1 -> value v2 -> value (thenEvt v1 v2)
| value_choose : forall v1 v2, value v1 -> value v2 -> value (chooseEvt v1 v2).

Fixpoint fix_value (e : exp) : Prop :=
  match e with
    | (const c) => True
    | (lam x e1) => True
    | (loc l) => True
    | (chan k) => True
    | (alwaysEvt e1) => fix_value e1
    | neverEvt => True
    | (thenEvt e1 e2) => fix_value e1 /\ fix_value e2
    | (chooseEvt e1 e2) => fix_value e1 /\ fix_value e2
    | (sendEvt e1 e2) => fix_value e1 /\ fix_value e2
    | (recvEvt e2) => fix_value e2
    | _ => False
  end.

Fact not_value_variable :
  ~ value x1.

unfold not. intros. inversion H.

Qed.

Theorem fix_value_correct :
  forall e, value e <-> fix_value e.

intros e. split.
generalize dependent e.
induction e; intros value_e; simpl; inversion value_e; auto.
generalize dependent e.
induction e; intros fix_value_e; simpl in fix_value_e;
auto using value_const, value_lam, value_loc, value_chan, 
value_always, value_never, value_recv; inversion fix_value_e;
destruct fix_value_e; auto using value_then, value_choose, value_send.

Qed.

Corollary use_fix_value :
  forall e, fix_value e -> value e.

intros e. assert (value e <-> fix_value e) as equiv.
apply fix_value_correct. destruct equiv. auto.

Qed.

Fact test_value :
  value (chooseEvt 
    (sendEvt k1 c1) 
    (thenEvt 
      (recvEvt k2) 
      (lam x1 (alwaysEvt x1)))).

apply use_fix_value. simpl. auto.

Qed.

Inductive ctxt : Set :=
  ctxt_hole : ctxt
| ctxt_app_l : ctxt -> exp -> ctxt
| ctxt_app_r : exp -> ctxt -> ctxt
| ctxt_ref : ctxt -> ctxt
| ctxt_get : ctxt -> ctxt
| ctxt_set_l : ctxt -> exp -> ctxt
| ctxt_set_r : exp -> ctxt -> ctxt
| ctxt_sync : ctxt -> ctxt
| ctxt_always : ctxt -> ctxt
| ctxt_then_l : ctxt -> exp -> ctxt
| ctxt_then_r : exp -> ctxt -> ctxt
| ctxt_choose_l : ctxt -> exp -> ctxt
| ctxt_choose_r : exp -> ctxt -> ctxt
| ctxt_send_l : ctxt -> exp -> ctxt
| ctxt_send_r : exp -> ctxt -> ctxt
| ctxt_recv : ctxt -> ctxt.

Fixpoint ctxt_fill_hole (M : ctxt) (e : exp) {struct M} : exp :=
  match M with
    | ctxt_hole => e
    | (ctxt_app_l M' e') => app (ctxt_fill_hole M' e) e'
    | (ctxt_app_r e' M') => app e' (ctxt_fill_hole M' e)
    | (ctxt_ref M') => ref (ctxt_fill_hole M' e)
    | (ctxt_get M') => get (ctxt_fill_hole M' e)
    | (ctxt_set_l M' e') => set (ctxt_fill_hole M' e) e'
    | (ctxt_set_r e' M') => set e' (ctxt_fill_hole M' e)
    | (ctxt_sync M') => sync (ctxt_fill_hole M' e)
    | (ctxt_always M') => alwaysEvt (ctxt_fill_hole M' e)
    | (ctxt_then_l M' e') => thenEvt (ctxt_fill_hole M' e) e'
    | (ctxt_then_r e' M') => thenEvt e' (ctxt_fill_hole M' e)
    | (ctxt_choose_l M' e') => chooseEvt (ctxt_fill_hole M' e) e'
    | (ctxt_choose_r e' M') => chooseEvt e' (ctxt_fill_hole M' e)
    | (ctxt_send_l M' e') => sendEvt (ctxt_fill_hole M' e) e'
    | (ctxt_send_r e' M') => sendEvt e' (ctxt_fill_hole M' e)
    | (ctxt_recv M') => recvEvt (ctxt_fill_hole M' e)
  end.

Notation "M [ e ]" := (ctxt_fill_hole M e) (at level 68).

Inductive get_ctxt : exp -> ctxt -> exp -> Prop :=

| get_ctxt_app_l : forall e1 e2 ctxt1 e1',
  ~ value e1 -> get_ctxt e1 ctxt1 e1' ->
  get_ctxt (app e1 e2) (ctxt_app_l ctxt1 e2) e1'

| get_ctxt_app_r : forall v1 e2 ctxt2 e2',
  value v1 -> ~ value e2 -> get_ctxt e2 ctxt2 e2' ->
  get_ctxt (app v1 e2) (ctxt_app_r v1 ctxt2) e2'

| get_ctxt_app : forall v1 v2,
  value v1 -> value v2 ->
  get_ctxt (app v1 v2) ctxt_hole (app v1 v2)

| get_ctxt_ref_under : forall e e' ctxt, 
  ~ value e -> get_ctxt e ctxt e' -> 
  get_ctxt (ref e) (ctxt_ref ctxt) e'

| get_ctxt_ref : forall v, 
  value v -> 
  get_ctxt (ref v) ctxt_hole (ref v)

| get_ctxt_get_under : forall e e' ctxt, 
  ~ value e -> get_ctxt e ctxt e' -> 
  get_ctxt (get e) (ctxt_get ctxt) e'

| get_ctxt_get : forall v, 
  value v -> 
  get_ctxt (get v) ctxt_hole (get v)

| get_ctxt_set_l : forall e1 e2 ctxt1 e1',
  ~ value e1 -> get_ctxt e1 ctxt1 e1' ->
  get_ctxt (set e1 e2) (ctxt_set_l ctxt1 e2) e1'

| get_ctxt_set_r : forall v1 e2 ctxt2 e2',
  value v1 -> ~ value e2 -> get_ctxt e2 ctxt2 e2' ->
  get_ctxt (set v1 e2) (ctxt_set_r v1 ctxt2) e2'

| get_ctxt_create : forall e,
  get_ctxt (create e) ctxt_hole (create e)

| get_ctxt_sync_under : forall e e' ctxt, 
  ~ value e -> get_ctxt e ctxt e' -> 
  get_ctxt (sync e) (ctxt_sync ctxt) e'

| get_ctxt_sync : forall v,
  value v ->
  get_ctxt (sync v) ctxt_hole (sync v)

| get_ctxt_newChan :
  get_ctxt newChan ctxt_hole newChan

| get_ctxt_always : forall e e' ctxt, 
  ~ value e -> get_ctxt e ctxt e' -> 
  get_ctxt (alwaysEvt e) (ctxt_always ctxt) e'

| get_ctxt_then_l : forall e1 e2 ctxt1 e1',
  ~ value e1 -> get_ctxt e1 ctxt1 e1' ->
  get_ctxt (thenEvt e1 e2) (ctxt_then_l ctxt1 e2) e1'

| get_ctxt_thenEvt_r : forall e1 e2 ctxt2 e2',
  value e1 -> ~ value e2 -> get_ctxt e2 ctxt2 e2' ->
  get_ctxt (thenEvt e1 e2) (ctxt_then_r e1 ctxt2) e2'

| get_ctxt_choose_l : forall e1 e2 ctxt1 e1',
  ~ value e1 -> get_ctxt e1 ctxt1 e1' ->
  get_ctxt (chooseEvt e1 e2) (ctxt_choose_l ctxt1 e2) e1'

| get_ctxt_choose_r : forall e1 e2 ctxt2 e2',
  value e1 -> ~ value e2 -> get_ctxt e2 ctxt2 e2' ->
  get_ctxt (chooseEvt e1 e2) (ctxt_choose_r e1 ctxt2) e2'

| get_ctxt_send_l : forall e1 e2 ctxt1 e1',
  ~ value e1 -> get_ctxt e1 ctxt1 e1' ->
  get_ctxt (sendEvt e1 e2) (ctxt_send_l ctxt1 e2) e1'

| get_ctxt_sendEvt_r : forall e1 e2 ctxt2 e2',
  value e1 -> ~ value e2 -> get_ctxt e2 ctxt2 e2' ->
  get_ctxt (sendEvt e1 e2) (ctxt_send_r e1 ctxt2) e2'

| get_ctxt_recv : forall e e' ctxt, 
  ~ value e -> get_ctxt e ctxt e' -> 
  get_ctxt (recvEvt e) (ctxt_recv ctxt) e'.

Lemma get_ctxt_unique : forall e M1 M2 e1 e2,
  get_ctxt e M1 e1 ->
  get_ctxt e M2 e2 ->
  (M1 = M2 /\ e1 = e2).

intros e M1 M2 e1 e2 get1 get2.

generalize dependent M1.
generalize dependent M2.

induction e; intros M2 get2 M1 get1.
inversion get1.
inversion get1.
inversion get1.
inversion get1; subst;
inversion get2; subst; try tauto.

assert (ctxt1 = ctxt0 /\ e1 = e2).
apply IHe1 with (M1 := ctxt1) (M2 := ctxt0).
apply H6. apply H4. destruct H. split. subst ctxt1. reflexivity. apply H0.

assert (ctxt2 = ctxt0 /\ e1 = e2).
apply IHe2 with (M1 := ctxt2) (M2 := ctxt0).
apply H8. apply H5. destruct H. split. subst ctxt2. reflexivity. apply H0.

inversion get1; subst;
inversion get2; subst; try tauto.

assert (ctxt0 = ctxt1 /\ e1 = e2).
apply IHe. apply H3. apply H1. destruct H. split. subst ctxt1. reflexivity. apply H4.

inversion get1; subst;
inversion get2; subst; try tauto.

assert (ctxt0 = ctxt1 /\ e1 = e2).
apply IHe. apply H3. apply H1. destruct H. split. subst ctxt1. reflexivity. apply H4.

inversion get1; subst;
inversion get2; subst; try tauto.

assert (ctxt1 = ctxt0 /\ e1 = e2).
apply IHe1 with (M1 := ctxt1) (M2 := ctxt0).
apply H6. apply H4. destruct H. split. subst ctxt1. reflexivity. apply H0.

assert (ctxt2 = ctxt0 /\ e1 = e2).
apply IHe2 with (M1 := ctxt2) (M2 := ctxt0).
apply H8. apply H5. destruct H. split. subst ctxt2. reflexivity. apply H0.

inversion get1.

inversion get1; subst;
inversion get2; subst;
auto.

inversion get1; subst;
inversion get2; subst;
try tauto.

assert (ctxt0 = ctxt1 /\ e1 = e2).
apply IHe. apply H3. apply H1. destruct H. split. subst ctxt1. reflexivity. apply H4.

inversion get1. inversion get2. auto.

inversion get1.

inversion get1; subst;
inversion get2; subst;
try tauto.

assert (ctxt0 = ctxt1 /\ e1 = e2).
apply IHe. apply H3. apply H1. destruct H. split. subst ctxt1. reflexivity. apply H4.

inversion get1.

inversion get1; subst;
inversion get2; subst; try tauto.

assert (ctxt1 = ctxt0 /\ e1 = e2).
apply IHe1 with (M1 := ctxt1) (M2 := ctxt0).
apply H6. apply H4. destruct H. split. subst ctxt1. reflexivity. apply H0.

assert (ctxt2 = ctxt0 /\ e1 = e2).
apply IHe2 with (M1 := ctxt2) (M2 := ctxt0).
apply H8. apply H5. destruct H. split. subst ctxt2. reflexivity. apply H0.

inversion get1; subst;
inversion get2; subst; try tauto.

assert (ctxt1 = ctxt0 /\ e1 = e2).
apply IHe1 with (M1 := ctxt1) (M2 := ctxt0).
apply H6. apply H4. destruct H. split. subst ctxt1. reflexivity. apply H0.

assert (ctxt2 = ctxt0 /\ e1 = e2).
apply IHe2 with (M1 := ctxt2) (M2 := ctxt0).
apply H8. apply H5. destruct H. split. subst ctxt2. reflexivity. apply H0.

inversion get1; subst;
inversion get2; subst; try tauto.

assert (ctxt1 = ctxt0 /\ e1 = e2).
apply IHe1 with (M1 := ctxt1) (M2 := ctxt0).
apply H6. apply H4. destruct H. split. subst ctxt1. reflexivity. apply H0.

assert (ctxt2 = ctxt0 /\ e1 = e2).
apply IHe2 with (M1 := ctxt2) (M2 := ctxt0).
apply H8. apply H5. destruct H. split. subst ctxt2. reflexivity. apply H0.

inversion get1; subst;
inversion get2; subst;
try tauto.

assert (ctxt0 = ctxt1 /\ e1 = e2).
apply IHe. apply H3. apply H1. destruct H. split. subst ctxt1. reflexivity. apply H4.

Qed.

Lemma get_ctxt_sync_value : forall e M v,
  get_ctxt e M (sync v) -> value v.

intros. 

generalize dependent e.
generalize dependent M.

induction M; intros; inversion H; subst; eauto.

Qed.

Fact test_fill_hole :
  forall e, ctxt_fill_hole ctxt_hole e = e.
  intros.
  simpl.
  reflexivity.
Qed.

Fact fill_hole_same :
  forall e ctxt e', get_ctxt e ctxt e' -> ctxt [ e' ] = e.

intros e1 ctxt1 e1' get_ctxt_e1_ctxt1_e1'.
induction get_ctxt_e1_ctxt1_e1'; subst; simpl; auto.

Qed.

Inductive evt_ctxt : Set :=
| evt_ctxt_hole : evt_ctxt
| evt_ctxt_then : evt_ctxt -> exp -> evt_ctxt.

Inductive get_evt_ctxt : exp -> evt_ctxt -> exp -> Prop :=

| get_evt_ctxt_choose : forall v1 v2,
  value v1 -> value v2 ->
  get_evt_ctxt (chooseEvt v1 v2) evt_ctxt_hole (chooseEvt v1 v2)

| get_evt_ctxt_send : forall (k:channel) v,
  value v ->
  get_evt_ctxt (sendEvt k v) evt_ctxt_hole (sendEvt k v)

| get_evt_ctxt_recv : forall (k:channel),
  get_evt_ctxt (recvEvt k) evt_ctxt_hole (recvEvt k)

| get_evt_ctxt_never :
  get_evt_ctxt neverEvt evt_ctxt_hole neverEvt

| get_evt_ctxt_then_always : forall v1 v2,
  value v1 -> value v2 ->
  get_evt_ctxt (thenEvt (alwaysEvt v1) v2) 
  evt_ctxt_hole (thenEvt (alwaysEvt v1) v2)

| get_evt_ctxt_then : forall v1 v2 v1' E,
  value v2 ->
  get_evt_ctxt v1 E v1' ->
  get_evt_ctxt (thenEvt v1 v2) (evt_ctxt_then E v2) v1'.

Fact evt_ctxt_only_values : forall e E e',
  get_evt_ctxt e E e' ->
  value e /\ value e'.

intros. 
split.

induction H;
auto using value_choose, value_send, value_recv, value_never, 
  value_then, value_always, value_chan.

induction H;
auto using value_choose, value_send, value_recv, value_never, 
  value_then, value_always, value_chan.
Qed.

Fact get_evt_ctxt_unique : forall e1 e2 e E1 E2,
  get_evt_ctxt e E1 e1 ->
  get_evt_ctxt e E2 e2 ->
  (E1 = E2 /\ e1 = e2).

intros e1 e2 e E1 E2 get1 get2.

generalize dependent e.
generalize dependent E2.

induction E1 as [ | E1 IH1 v1 ].

intros E2 e get1 get2.
inversion get1; subst; inversion get2; subst; auto.
inversion H6.

induction E2 as [ | E2 IH2 v2 ]. 

intros e get1 get2.
inversion get1; subst; inversion get2; subst; auto.
inversion H4.

intros e get1 get2.
inversion get1; subst; inversion get2; subst.
assert (E1 = E2 /\ e1 = e2) as ind_step.
apply IH1 with (E2 := E2) (e := v0). auto. auto.
destruct ind_step. subst E2 e2.
split. reflexivity. reflexivity.

Qed.

Fixpoint evt_fill_hole (E : evt_ctxt) (e : exp) {struct E} : exp :=
  match E with
    | (evt_ctxt_then E' e') => thenEvt (evt_fill_hole E' e) e'
    | evt_ctxt_hole => e
  end.

Axiom heap : Set.
Axiom fresh_heap : heap.
Axiom st_step : heap -> exp -> heap -> exp -> Prop.

(* Notation "H ; e ->st H' ; e'" := (st_step H e H' e') (at level 65). *)

Axiom tid : Set.
Axiom id1 id2 id3 : tid.
Definition mt_thread := prod tid exp.
Axiom mt_step : 
  heap -> Ensemble mt_thread -> 
  heap -> Ensemble mt_thread -> Prop.

Definition mt_unique_tids (T:Ensemble mt_thread) : Prop :=
  forall t1 t2 id1 id2 e1 e2,
    t1 = (id1, e1) -> t2 = (id2, e2) ->
    In mt_thread T t1 -> In mt_thread T t2 ->
    t1 <> t2 -> id1 <> id2.

Axiom mt_step_preserves_unique_tids : 
  forall H T H' T',
    mt_step H T H' T' ->
    mt_unique_tids T ->
    mt_unique_tids T'.

Inductive ldvalue : list ctxt -> exp -> Prop :=

| ldvalue_send : forall list_M v1 v2 E e,
  value v1 -> value v2 ->
  get_evt_ctxt e E (sendEvt v1 v2) ->
  ldvalue list_M e

| ldvalue_recv : forall list_M v E e,
  value v -> 
  get_evt_ctxt e E (recvEvt v) ->
  ldvalue list_M e

| ldvalue_always : forall v,
  value v ->
  ldvalue nil (alwaysEvt v).


(* Notation "H ; T ->mt H' ; T'" := (mt_step H T H' T') (at level 65). *)