Require Import Coq.Logic.Classical.
Require Import Coq.Arith.Arith.
Require Import Coq.Sets.Ensembles.
Require Import syntax.

Axiom nonce : Set.

Inductive path_elt : Set :=
| path_left : path_elt
| path_right : path_elt
| path_send : nonce -> tid -> path -> path_elt
| path_recv : nonce -> tid -> path -> path_elt
| path_heap : nonce -> heap_path -> path_elt
with heap_path_elt : Set :=
| heap_path_thread : nonce -> tid -> path -> heap_path_elt
with path : Set :=
| path_nil : path
| path_cons : path_elt -> path -> path
with heap_path : Set :=
| heap_path_nil : heap_path
| heap_path_cons : heap_path_elt -> heap_path -> heap_path.

Fact paths_are_finite :
  forall elt p, p <> path_cons elt p.

intros elt p.
unfold not.
intros false.
induction p.
inversion false.
apply IHp.
inversion false. subst p. apply false.

Qed.

Fact heap_paths_are_finite :
  forall elt hp, hp <> heap_path_cons elt hp.

intros elt hp.
unfold not.
intros false.
induction hp as [ | elt' hp' ].
inversion false.
apply IHhp'.
inversion false as [ [ H1 H2 ] ].
rewrite <- H2.
apply H2.

Qed.

Scheme path_elt_induction := Induction for path_elt Sort Prop
with heap_path_elt_induction := Induction for heap_path_elt Sort Prop
with path_induction := Induction for path Sort Prop
with heap_path_induction := Induction for heap_path Sort Prop.

Combined Scheme trail_induction from path_induction, heap_path_induction.

Inductive trail : Set :=
| thread_trail : tid -> path -> trail
| heap_trail : heap_path -> trail.

Inductive extends : path -> path -> Prop :=
| extends_refl : forall p, extends p p
| extends_cons : forall elt p1 p2,
  extends p1 p2 ->
  extends (path_cons elt p1) p2.

Inductive extends_strict : path -> path -> Prop :=
| extends_strict_cons : forall p1 p2 elt p1',
  p1 = path_cons elt p1' ->
  extends p1' p2 ->
  extends_strict p1 p2.

Inductive heap_extends : heap_path -> heap_path -> Prop :=
| heap_extends_refl : forall hp, heap_extends hp hp
| heap_extends_cons : forall elt hp1 hp2,
  heap_extends hp1 hp2 ->
  heap_extends (heap_path_cons elt hp1) hp2.

Inductive heap_extends_strict : heap_path -> heap_path -> Prop :=
| heap_extends_strict_cons : forall hp1 hp2 elt hp1',
  hp1 = heap_path_cons elt hp1' ->
  heap_extends hp1' hp2 ->
  heap_extends_strict hp1 hp2.

Fixpoint path_length (p:path) : nat :=
  match p with
    | path_nil => 0
    | path_cons elt p' => 1 + path_length p'
  end.

Fixpoint heap_path_length (hp:heap_path) : nat :=
  match hp with
    | heap_path_nil => 0
    | heap_path_cons elt hp' => 1 + heap_path_length hp'
  end.

Lemma extends_implies_length_ge :
  forall p1 p2,
    extends p1 p2 ->
    path_length p1 >= path_length p2.

intros p1 p2 p1_extends_p2.
induction p1_extends_p2.
auto with arith.
simpl. auto with arith.

Qed.

Lemma heap_extends_implies_length_ge :
  forall hp1 hp2,
    heap_extends hp1 hp2 ->
    heap_path_length hp1 >= heap_path_length hp2.

intros hp1 hp2 hp1_extends_hp2.
induction hp1_extends_hp2.
auto with arith.
simpl. auto with arith.

Qed.

Lemma extends_strict_implies_length_gt : 
  forall p1 p2,
    extends_strict p1 p2 ->
    path_length p1 > path_length p2.

intros p1 p2 strict.
inversion strict as [ _p1 _p2 elt p1' p1_eq p1'_extends_p2 ].
subst p1 _p1 _p2.
assert (path_length p1' >= path_length p2) as p1'_ge_p2.
apply extends_implies_length_ge. auto.
simpl. auto with arith.

Qed.

Lemma heap_extends_strict_implies_length_gt : 
  forall hp1 hp2,
    heap_extends_strict hp1 hp2 ->
    heap_path_length hp1 > heap_path_length hp2.

intros hp1 hp2 strict.
inversion strict as [ _hp1 _hp2 elt hp1' hp1_eq hp1'_extends_hp2 ].
subst hp1 _hp1 _hp2.
assert (heap_path_length hp1' >= heap_path_length hp2) as hp1'_ge_hp2.
apply heap_extends_implies_length_ge. auto.
simpl. auto with arith.

Qed.

Lemma extends_strict_implies_ne :
  forall p1 p2,
    extends_strict p1 p2 ->
    p1 <> p2.

intros p1 p2 strict eq. subst p2.
assert (path_length p1 > path_length p1).
apply extends_strict_implies_length_gt. auto.
assert (~ path_length p1 > path_length p1).
apply gt_irrefl.
tauto.

Qed.

Lemma heap_extends_strict_implies_ne :
  forall hp1 hp2,
    heap_extends_strict hp1 hp2 ->
    hp1 <> hp2.

intros hp1 hp2 strict eq. subst hp2.
assert (heap_path_length hp1 > heap_path_length hp1).
apply heap_extends_strict_implies_length_gt. auto.
assert (~ heap_path_length hp1 > heap_path_length hp1).
apply gt_irrefl.
tauto.

Qed.

Lemma extends_strict_implies_extends :
  forall p1 p2,
    extends_strict p1 p2 ->
    extends p1 p2.

intros p1 p2 strict.
inversion strict.
subst p1.
eapply extends_cons.
auto.

Qed.

Lemma heap_extends_strict_implies_extends :
  forall hp1 hp2,
    heap_extends_strict hp1 hp2 ->
    heap_extends hp1 hp2.

intros hp1 hp2 strict.
inversion strict.
subst hp1.
eapply heap_extends_cons.
auto.

Qed.

Lemma extends_strict_implies_reverse_not_extends :
  forall p1 p2,
    extends_strict p1 p2 ->
    ~ (extends p2 p1).

unfold not. intros p1 p2 strict ext.
assert (path_length p1 > path_length p2) as gt1.
apply extends_strict_implies_length_gt. auto.
assert (path_length p2 >= path_length p1) as ge1.
apply extends_implies_length_ge. auto.
unfold ge in ge1.
assert (~ path_length p1 > path_length p2).
apply le_not_gt. auto. tauto.

Qed.

Lemma heap_extends_strict_implies_reverse_not_extends :
  forall hp1 hp2,
    heap_extends_strict hp1 hp2 ->
    ~ (heap_extends hp2 hp1).

unfold not. intros hp1 hp2 strict ext.
assert (heap_path_length hp1 > heap_path_length hp2) as gt1.
apply heap_extends_strict_implies_length_gt. auto.
assert (heap_path_length hp2 >= heap_path_length hp1) as ge1.
apply heap_extends_implies_length_ge. auto.
unfold ge in ge1.
assert (~ heap_path_length hp1 > heap_path_length hp2).
apply le_not_gt. auto. tauto.

Qed.

Lemma extends_trans :
  forall p1 p2 p3,
    extends p1 p2 ->
    extends p2 p3 ->
    extends p1 p3.

intros p1 p2 p3 extends12 extends23.

induction extends12.
auto.
apply extends_cons. auto.

Qed.

Lemma heap_extends_trans :
  forall hp1 hp2 hp3,
    heap_extends hp1 hp2 ->
    heap_extends hp2 hp3 ->
    heap_extends hp1 hp3.

intros hp1 hp2 hp3 extends12 extends23.

induction extends12.
auto.
apply heap_extends_cons. auto.

Qed.

Lemma nil_extends_only_nil :
  forall p,
    extends path_nil p ->
    p = path_nil.

  intros p extends_p.
  inversion extends_p.
  auto.

Qed.

Lemma nil_heap_extends_only_nil :
  forall p,
    heap_extends heap_path_nil p ->
    p = heap_path_nil.

  intros p extends_p.
  inversion extends_p.
  auto.

Qed.

Lemma common_parent_implies_extends :
  forall p p1 p2,
    extends p p1 ->
    extends p p2 ->
    extends p1 p2 \/ extends p2 p1.

  intros p p1 p2 H1 H2.

  (* Proof by induction of the length of the common parent *)

  generalize dependent p.
  induction p as [ | elt p IH ].

  (* Case 1: parent = nil *)
  intros H1 H2.
  assert (p1 = path_nil). apply nil_extends_only_nil. auto.
  assert (p2 = path_nil). apply nil_extends_only_nil. auto.
  subst p1 p2. left. apply extends_refl.

  (* Case 2: parent = elt : parent' *)
  intros H1 H2.
  remember (path_cons elt p) as p'.
  inversion H1 as [ p1' | _elt1 _p1 p1_ p_extends_p1 t'_eq1 ];
  inversion H2 as [ p2' | _elt2 _p2 p2_ p_extends_p2 t'_eq2 ].

    (* Case 2a: parent = p1 = p2 *)
    subst p1 p2. left. apply extends_refl.

    (* Case 2b: parent = p1 and parent strictly extends p2 *)
    subst p1' p' p2_ p1. inversion t'_eq2. subst _elt2 _p2.
    left. apply extends_cons. auto.

    (* Case 2b: parent = p2 and parent strictly extends p1 *)
    subst p2' p' p1_ p2. inversion t'_eq1. subst _elt1 _p1.
    right. apply extends_cons. auto.

    (* Case 2a: parent strictly extends both p1 and l2 *)
    subst p1_ p2_ p'. inversion t'_eq1. inversion t'_eq2.
    subst _elt1 _p1 _elt2 _p2. auto.

Qed.

Lemma common_parent_implies_heap_extends :
  forall p p1 p2,
    heap_extends p p1 ->
    heap_extends p p2 ->
    heap_extends p1 p2 \/ heap_extends p2 p1.

  intros p p1 p2 H1 H2.

  (* Proof by induction of the length of the common parent *)

  generalize dependent p.
  induction p as [ | elt p IH ].

  (* Case 1: parent = nil *)
  intros H1 H2.
  assert (p1 = heap_path_nil). apply nil_heap_extends_only_nil. auto.
  assert (p2 = heap_path_nil). apply nil_heap_extends_only_nil. auto.
  subst p1 p2. left. apply heap_extends_refl.

  (* Case 2: parent = elt : parent' *)
  intros H1 H2.
  remember (heap_path_cons elt p) as p'.
  inversion H1 as [ p1' | _elt1 _p1 p1_ p_heap_extends_p1 t'_eq1 ];
  inversion H2 as [ p2' | _elt2 _p2 p2_ p_heap_extends_p2 t'_eq2 ].

    (* Case 2a: parent = p1 = p2 *)
    subst p1 p2. left. apply heap_extends_refl.

    (* Case 2b: parent = p1 and parent strictly extends p2 *)
    subst p1' p' p2_ p1. inversion t'_eq2. subst _elt2 _p2.
    left. apply heap_extends_cons. auto.

    (* Case 2b: parent = p2 and parent strictly extends p1 *)
    subst p2' p' p1_ p2. inversion t'_eq1. subst _elt1 _p1.
    right. apply heap_extends_cons. auto.

    (* Case 2a: parent strictly extends both p1 and l2 *)
    subst p1_ p2_ p'. inversion t'_eq1. inversion t'_eq2.
    subst _elt1 _p1 _elt2 _p2. auto.

Qed.

Fixpoint elt_count (p:path) : nat :=
  match p with
    | path_nil => 0
    | path_cons elt p' =>
      match elt with
        | path_left => 1 + elt_count p'
        | path_right => 1 + elt_count p'
        | path_send n_x id_x p_x => 1 + elt_count p' + elt_count p_x
        | path_recv n_x id_x p_x => 1 + elt_count p' + elt_count p_x
        | path_heap n_x hp_x => 1 + elt_count p' + heap_elt_count hp_x
      end
  end
with heap_elt_count (hp:heap_path) : nat :=
  match hp with
    | heap_path_nil => 0
    | heap_path_cons elt hp' =>
      match elt with
        | heap_path_thread n_x id_x p_x => 1 + heap_elt_count hp' + elt_count p_x
      end
  end.

Definition trail_elt_count (t:trail) : nat :=
  match t with
    | thread_trail id p => elt_count p
    | heap_trail p => heap_elt_count p
  end.

Lemma extends_implies_elt_count_ge :
  forall p1 p2,
    extends p1 p2 ->
    elt_count p1 >= elt_count p2.

intros p1 p2 p1_extends_p2.
induction p1_extends_p2 as [ | elt ].
auto.
simpl. induction elt; auto with arith.

Qed.

Lemma heap_extends_implies_elt_count_ge :
  forall p1 p2,
    heap_extends p1 p2 ->
    heap_elt_count p1 >= heap_elt_count p2.

intros p1 p2 p1_extends_p2.
induction p1_extends_p2 as [ | elt ].
auto.
simpl. induction elt; auto with arith.

Qed.

Lemma extends_strict_implies_elt_count_gt :
  forall p1 p2,
    extends_strict p1 p2 ->
    elt_count p1 > elt_count p2.

intros p1 p2 strict.
inversion strict as [ _p1 _p2 elt p1' ].
assert (elt_count p1' >= elt_count p2).
apply extends_implies_elt_count_ge. auto.
subst p1. simpl. induction elt; auto with arith.

Qed.

Lemma heap_extends_strict_implies_elt_count_gt :
  forall p1 p2,
    heap_extends_strict p1 p2 ->
    heap_elt_count p1 > heap_elt_count p2.

intros p1 p2 strict.
inversion strict as [ _p1 _p2 elt p1' ].
assert (heap_elt_count p1' >= heap_elt_count p2).
apply heap_extends_implies_elt_count_ge. auto.
subst p1. simpl. induction elt; auto with arith.

Qed.

Lemma extends_cons_implies_extends_strict :
  forall p1 p2 elt,
    extends p1 (path_cons elt p2) ->
    extends_strict p1 p2.

intros p1 p2 elt ext.
assert (extends p1 p2) as ext2.
eapply extends_trans. apply ext. apply extends_cons. apply extends_refl.
inversion ext2 as [ | elt2 ].
subst p2. assert (path_length p1 >= 1 + path_length p1) as false.
assert (path_length (path_cons elt p1) = 1 + path_length p1) as rewriting.
simpl. reflexivity. rewrite <- rewriting.
apply extends_implies_length_ge. auto.
unfold ge in false.
assert (~ 1 + path_length p1 <= path_length p1).
auto with arith. tauto.

subst p1 p3. eapply extends_strict_cons. reflexivity. auto.

Qed.

Lemma heap_extends_cons_implies_extends_strict :
  forall p1 p2 elt,
    heap_extends p1 (heap_path_cons elt p2) ->
    heap_extends_strict p1 p2.

intros p1 p2 elt ext.
assert (heap_extends p1 p2) as ext2.
eapply heap_extends_trans. apply ext.
apply heap_extends_cons. apply heap_extends_refl.
inversion ext2 as [ | elt2 ].
subst p2. assert (heap_path_length p1 >= 1 + heap_path_length p1) as false.
assert (heap_path_length (heap_path_cons elt p1) = 
  1 + heap_path_length p1) as rewriting.
simpl. reflexivity. rewrite <- rewriting.
apply heap_extends_implies_length_ge. auto.
unfold ge in false.
assert (~ 1 + heap_path_length p1 <= heap_path_length p1).
auto with arith. tauto.

subst p1 p2. eapply heap_extends_strict_cons. reflexivity. auto.

Qed.
